@@ -1,14 +1,19 @@
-#This file is part of Tryton.  The COPYRIGHT file at the top level of
-#this repository contains the full copyright notices and license terms.
+# This file is part of Tryton.  The COPYRIGHT file at the top level of
+# this repository contains the full copyright notices and license terms.
 from decimal import Decimal
 from dateutil.relativedelta import relativedelta
+
+from sql import Null, Column
+
 from trytond.model import ModelView, ModelSQL, fields
 from trytond import backend
 from trytond.pyson import Eval
 from trytond.transaction import Transaction
 from trytond.pool import Pool
+from trytond.wizard import Wizard, StateView, Button
 
-__all__ = ['PaymentTerm', 'PaymentTermLine']
+__all__ = ['PaymentTerm', 'PaymentTermLine', 'PaymentTermLineRelativeDelta',
+    'TestPaymentTerm', 'TestPaymentTermView', 'TestPaymentTermViewResult']
 
 
 class PaymentTerm(ModelSQL, ModelView):
@@ -55,7 +60,7 @@ class PaymentTerm(ModelSQL, ModelView):
         If specified, date will be used as the start date, otherwise current
         date will be used.
         """
-        #TODO implement business_days
+        # TODO implement business_days
         # http://pypi.python.org/pypi/BusinessHours/
         Date = Pool().get('ir.date')
 
@@ -121,44 +126,13 @@ class PaymentTermLine(ModelSQL, ModelView):
             }, depends=['type'])
     currency_digits = fields.Function(fields.Integer('Currency Digits'),
         'on_change_with_currency_digits')
-    day = fields.Integer('Day of Month')
-    month = fields.Selection([
-            (None, ''),
-            ('1', 'January'),
-            ('2', 'February'),
-            ('3', 'March'),
-            ('4', 'April'),
-            ('5', 'May'),
-            ('6', 'June'),
-            ('7', 'July'),
-            ('8', 'August'),
-            ('9', 'September'),
-            ('10', 'October'),
-            ('11', 'November'),
-            ('12', 'December'),
-            ], 'Month', sort=False)
-    weekday = fields.Selection([
-            (None, ''),
-            ('0', 'Monday'),
-            ('1', 'Tuesday'),
-            ('2', 'Wednesday'),
-            ('3', 'Thursday'),
-            ('4', 'Friday'),
-            ('5', 'Saturday'),
-            ('6', 'Sunday'),
-            ], 'Day of Week', sort=False)
-    months = fields.Integer('Number of Months', required=True)
-    weeks = fields.Integer('Number of Weeks', required=True)
-    days = fields.Integer('Number of Days', required=True)
+    relativedeltas = fields.One2Many(
+        'account.invoice.payment_term.line.relativedelta', 'line', 'Deltas')
 
     @classmethod
     def __setup__(cls):
         super(PaymentTermLine, cls).__setup__()
         cls._order.insert(0, ('sequence', 'ASC'))
-        cls._sql_constraints += [
-            ('day', 'CHECK(day BETWEEN 1 AND 31)',
-                'Day of month must be between 1 and 31.'),
-            ]
         cls._error_messages.update({
                 'invalid_percentage_and_divisor': ('Percentage and '
                     'Divisor values are not consistent in line "%(line)s" '
@@ -201,7 +175,7 @@ class PaymentTermLine(ModelSQL, ModelView):
     @staticmethod
     def order_sequence(tables):
         table, _ = tables[None]
-        return [table.sequence == None, table.sequence]
+        return [table.sequence == Null, table.sequence]
 
     @staticmethod
     def default_currency_digits():
@@ -211,46 +185,30 @@ class PaymentTermLine(ModelSQL, ModelView):
     def default_type():
         return 'remainder'
 
-    @staticmethod
-    def default_months():
-        return 0
-
-    @staticmethod
-    def default_weeks():
-        return 0
-
-    @staticmethod
-    def default_days():
-        return 0
-
     @fields.depends('type')
     def on_change_type(self):
-        res = {}
         if self.type != 'fixed':
-            res['amount'] = Decimal('0.0')
-            res['currency'] = None
+            self.amount = Decimal('0.0')
+            self.currency = None
         if self.type not in ('percent', 'percent_on_total'):
-            res['percentage'] = Decimal('0.0')
-            res['divisor'] = Decimal('0.0')
-        return res
+            self.percentage = Decimal('0.0')
+            self.divisor = Decimal('0.0')
 
     @fields.depends('percentage')
     def on_change_percentage(self):
         if not self.percentage:
-            return {'divisor': 0.0}
-        return {
-            'divisor': self.round(Decimal('100.0') / self.percentage,
-                self.__class__.divisor.digits[1]),
-            }
+            self.divisor = 0.0
+        else:
+            self.divisor = self.round(Decimal('100.0') / self.percentage,
+                self.__class__.divisor.digits[1])
 
     @fields.depends('divisor')
     def on_change_divisor(self):
         if not self.divisor:
-            return {'percentage': 0.0}
-        return {
-            'percentage': self.round(Decimal('100.0') / self.divisor,
-                self.__class__.percentage.digits[1]),
-            }
+            self.percentage = 0.0
+        else:
+            self.percentage = self.round(Decimal('100.0') / self.divisor,
+                self.__class__.percentage.digits[1])
 
     @fields.depends('currency')
     def on_change_with_currency_digits(self, name=None):
@@ -269,7 +227,9 @@ class PaymentTermLine(ModelSQL, ModelView):
             }
 
     def get_date(self, date):
-        return date + relativedelta(**self.get_delta())
+        for relativedelta_ in self.relativedeltas:
+            date += relativedelta_.get()
+        return date
 
     def get_value(self, remainder, amount, currency):
         Currency = Pool().get('currency.currency')
@@ -323,3 +283,169 @@ class PaymentTermLine(ModelSQL, ModelView):
                         'line': line.rec_name,
                         'term': line.payment.rec_name,
                         })
+
+
+class PaymentTermLineRelativeDelta(ModelSQL, ModelView):
+    'Payment Term Line Relative Delta'
+    __name__ = 'account.invoice.payment_term.line.relativedelta'
+    sequence = fields.Integer('Sequence')
+    line = fields.Many2One('account.invoice.payment_term.line',
+        'Payment Term Line', required=True, ondelete='CASCADE')
+    day = fields.Integer('Day of Month',
+        domain=['OR',
+            ('day', '=', None),
+            [('day', '>=', 1), ('day', '<=', 31)],
+            ])
+    month = fields.Selection([
+            (None, ''),
+            ('1', 'January'),
+            ('2', 'February'),
+            ('3', 'March'),
+            ('4', 'April'),
+            ('5', 'May'),
+            ('6', 'June'),
+            ('7', 'July'),
+            ('8', 'August'),
+            ('9', 'September'),
+            ('10', 'October'),
+            ('11', 'November'),
+            ('12', 'December'),
+            ], 'Month', sort=False)
+    weekday = fields.Selection([
+            (None, ''),
+            ('0', 'Monday'),
+            ('1', 'Tuesday'),
+            ('2', 'Wednesday'),
+            ('3', 'Thursday'),
+            ('4', 'Friday'),
+            ('5', 'Saturday'),
+            ('6', 'Sunday'),
+            ], 'Day of Week', sort=False)
+    months = fields.Integer('Number of Months', required=True)
+    weeks = fields.Integer('Number of Weeks', required=True)
+    days = fields.Integer('Number of Days', required=True)
+
+    @classmethod
+    def __setup__(cls):
+        super(PaymentTermLineRelativeDelta, cls).__setup__()
+        cls._order.insert(0, ('sequence', 'ASC'))
+
+    @classmethod
+    def __register__(cls, module_name):
+        TableHandler = backend.get('TableHandler')
+        cursor = Transaction().cursor
+        pool = Pool()
+        Line = pool.get('account.invoice.payment_term.line')
+        sql_table = cls.__table__()
+        line = Line.__table__()
+
+        super(PaymentTermLineRelativeDelta, cls).__register__(module_name)
+
+        line_table = TableHandler(cursor, Line, module_name)
+
+        # Migration from 3.4
+        fields = ['day', 'month', 'weekday', 'months', 'weeks', 'days']
+        if any(line_table.column_exist(f) for f in fields):
+            columns = ([line.id.as_('line')]
+                + [Column(line, f) for f in fields])
+            cursor.execute(*sql_table.insert(
+                    columns=[sql_table.line]
+                    + [Column(sql_table, f) for f in fields],
+                    values=line.select(*columns)))
+            for field in fields:
+                line_table.drop_column(field, exception=True)
+
+    @staticmethod
+    def order_sequence(tables):
+        table, _ = tables[None]
+        return [table.sequence == Null, table.sequence]
+
+    @staticmethod
+    def default_months():
+        return 0
+
+    @staticmethod
+    def default_weeks():
+        return 0
+
+    @staticmethod
+    def default_days():
+        return 0
+
+    def get(self):
+        "Return the relativedelta"
+        return relativedelta(
+            day=self.day,
+            month=int(self.month) if self.month else None,
+            days=self.days,
+            weeks=self.weeks,
+            months=self.months,
+            weekday=int(self.weekday) if self.weekday else None,
+            )
+
+
+class TestPaymentTerm(Wizard):
+    'Test Payment Term'
+    __name__ = 'account.invoice.payment_term.test'
+    start_state = 'test'
+    test = StateView('account.invoice.payment_term.test',
+        'account_invoice.payment_term_test_view_form',
+        [Button('Close', 'end', 'tryton-close', default=True)])
+
+    def default_test(self, fields):
+        context = Transaction().context
+        default = {}
+        if context['active_model'] == 'account.invoice.payment_term':
+            default['payment_term'] = context['active_id']
+        return default
+
+
+class TestPaymentTermView(ModelView):
+    'Test Payment Term'
+    __name__ = 'account.invoice.payment_term.test'
+    payment_term = fields.Many2One('account.invoice.payment_term',
+        'Payment Term', required=True)
+    date = fields.Date('Date')
+    amount = fields.Numeric('Amount', required=True,
+        digits=(16, Eval('currency_digits', 2)), depends=['currency_digits'])
+    currency = fields.Many2One('currency.currency', 'Currency', required=True)
+    currency_digits = fields.Integer('Currency Digits')
+    result = fields.One2Many('account.invoice.payment_term.test.result',
+        None, 'Result', readonly=True)
+
+    @staticmethod
+    def default_currency():
+        pool = Pool()
+        Company = pool.get('company.company')
+        company = Transaction().context.get('company')
+        if company:
+            return Company(company).currency.id
+
+    @fields.depends('currency')
+    def on_change_with_currency_digits(self):
+        if self.currency:
+            return self.currency.digits
+        return 2
+
+    @fields.depends('payment_term', 'date', 'amount', 'currency', 'result')
+    def on_change_with_result(self):
+        pool = Pool()
+        Result = pool.get('account.invoice.payment_term.test.result')
+        result = []
+        if (self.payment_term and self.amount and self.currency):
+            for date, amount in self.payment_term.compute(
+                    self.amount, self.currency, self.date):
+                result.append(Result(
+                        date=date, amount=amount,
+                        currency_digits=self.currency.digits))
+        self.result = result
+        return self._changed_values.get('result', [])
+
+
+class TestPaymentTermViewResult(ModelView):
+    'Test Payment Term'
+    __name__ = 'account.invoice.payment_term.test.result'
+    date = fields.Date('Date', readonly=True)
+    amount = fields.Numeric('Amount', readonly=True,
+        digits=(16, Eval('currency_digits', 2)), depends=['currency_digits'])
+    currency_digits = fields.Integer('Currency Digits')
